;;; emacsia-level.el --- Level manipulation for Emacsia

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords: 

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;


;;; Code:

(defstruct emacsia-level  
  name					; name of the level
  depth					; depth
  map					; the map
  transmap				; the opaqueness map (non-nil
					; means opaque)
  visible)				; the parts of the level that
					; have been lighted  

(defvar emacsia-level-current nil
  "The current level")

(defun emacsia-set-level (level)
  "Change the current level to LEVEL."
  ;; run hooks
  ;; fix level
  (setq emacsia-level-current level))

(defun emacsia-level-get-height (level)
  "Return the height of the LEVEL."
  (length (emacsia-level-map level)))

(defun emacsia-level-get-width (level)
  "Return the width of the LEVEL."
  (length (aref (emacsia-level-map level) 0)))

(defun emacsia-level-pos-valid-p (pos)
  "Return non-nil if POS is a valid coordinate."
  (and (emacsia-pos-p pos)
       (>= (emacsia-pos-x pos) 0)
       (>= (emacsia-pos-y pos) 0)
       (< (emacsia-pos-y pos)
	  (emacsia-level-get-height emacsia-level-current))
       (< (emacsia-pos-x pos)
	  (emacsia-level-get-width emacsia-level-current))))


(defun emacsia-level-get-block (pos)
  "Return the block at POS in `emacsia-level-current'."
  (aref emacsia-block-vector
	(aref (aref (emacsia-level-map emacsia-level-current)
		    (emacsia-pos-y pos))
	      (emacsia-pos-x pos))))

;; (defun emacsia-level-generate-transmap (level)
;;   "Generate the transparancy map for LEVEL."
;;   (let ((height (emacsia-level-get-height level))
;; 	(width (emacsia-level-get-width level))
;; 	(transmap (emacsia-level-transmap level))
;; 	;; hack to allow for easy getblock
;; 	(emacsia-level-current level))
;;     (unless transmap
;;       (setq transmap (make-vector height nil))
;;       (dotimes (i height)
;; 	(aset transmap i (make-vector width nil))))
;;     (dotimes (y height)
;;       (dotimes (x width)
;; 	(aset (aref transmap y) x
;; 	      (not (emacsia-block-opaque
;; 		    ;; get the block!
;; 		    (emacsia-level-get-block (emacsia-pos x y)))))))
;;     (setf (emacsia-level-transmap level) transmap)
;;     level))

(defun emacsia-level-set-block (pos block)
  "Set the block at POS in `emacsia-level-current' to BLOCK.

If this block is visible, change it in the visible map as well."
  (aset (aref (emacsia-level-map emacsia-level-current)
	      (emacsia-pos-y pos))
	(emacsia-pos-x pos) (emacsia-block-index block)))

;; also update the transparency map
; (aref (aref (emacsia-level-transmap emacsia-level-current)
;; 		    (emacsia-pos-y pos))
;; 	      (emacsia-pos-x pos))
;; 	(emacsia-block-opaque block)

;; (when (not (equal (emacsia-level-get-visible-block pos)
;; 		    emacsia-block-unexamined))
;;     ;; This block has been examined, so we should update it both
;;     ;; places.
;;     (emacsia-level-set-visible-block pos block))

(defun emacsia-level-set-visible-block (pos block)
  "Change what's visible at POS in the current level to BLOCK."
  ;; if a game has started, this calls for a UI-update.
  (when emacsia-playing-p
    (emacsia-touch pos))
  (aset (aref (emacsia-level-visible emacsia-level-current)
	      (emacsia-pos-y pos))
	(emacsia-pos-x pos)
	(emacsia-block-index block)))

(defun emacsia-level-get-visible-block (pos)
  "Return the visible block at  POS in the current level."
  ;; if a game has started, this requires a UI-update.
  (when emacsia-playing-p
    (emacsia-touch pos))
  (aref emacsia-block-vector
	(aref (aref (emacsia-level-visible emacsia-level-current)
		    (emacsia-pos-y pos))
	      (emacsia-pos-x pos))))

(defun emacsia-level-light (pos)
  "Light a position."
  ;; light the tile beneath
  (emacsia-level-set-visible-block pos (emacsia-level-get-block pos))
  (emacsia-touch pos))

(defun emacsia-level-unlight (pos)
  "Unlight a position."
  (unless (emacsia-block-permalit (emacsia-level-get-visible-block pos))
    (emacsia-level-set-visible-block pos emacsia-block-unexamined))
  (emacsia-touch pos))

(defun emacsia-level-light-radius (pos radius)
  "Apply lighting at POS in RADIUS."
  (let ((coords (emacsia-pos-expand-radius pos radius)))
    (mapc 'emacsia-level-light coords)))

(defun emacsia-level-unlight-radius (pos radius)
  "Apply unlighting at POS in RADIUS."
  (let ((coords (emacsia-pos-expand-radius pos radius)))
    (mapc 'emacsia-level-unlight coords)))

(provide 'emacsia-level)
;;; emacsia-level.el ends here
