;;; emacsia-los.el --- Line of Sight code

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords: 

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This module should handle LOS-code for Emcasia.

;;; Code:

(defvar emacsia-los-max-range 11
  "The maximum range the players LOS lasts.")

(defvar emacsia-max-range-ring nil
  "A list of relative coordinates, constituting a ring in the
  max-range.")

(defun emacsia-los-slope (pos1 pos2)
  "Return the slop of a line from POS1 to POS2."
  (/ (- (emacsia-pos-x pos1) (emacsia-pos-x pos2))
     (- (emacsia-pos-y pos1) (emacsia-pos-y pos2))))

(defun emacsia-los-invslope (pos1 pos2)
  (/ 1 (emacsia-los-slope pos1 pos2)))

(defun emacsia-los-generate-range-ring ()
  "Generate `emacsia-max-range-ring'."
  (let ((x (- emacsia-los-max-range))
	(results nil))
    (while (< x emacsia-los-max-range)
      (let ((y (- emacsia-los-max-range)))
	(while (< y emacsia-los-max-range)
	  (setq results (append results (list (emacsia-pos x y))))
	  (setq y (+ y 1))))
      (setq x (+ x 1)))
    ;; remove all fields not satisfying the distance-requirement
    (let ((orego (emacsia-pos 0 0)))
      (remove-if-not
       (lambda (pos)
	 (and 
	  (emacsia-pos-distance-predicate-p
	   orego
	   pos
	   (+ emacsia-los-max-range 2) '<)
	  (emacsia-pos-distance-predicate-p
	   orego
	   pos
	   emacsia-los-max-range '>=)))

       results))))

(defun emacsia-los-get-blockers (coordinates pos)
  "Return all elements of COORDINATES that block LOS.

The returned list is sorted by how far the element is from POS."
  ;; (sort 
  (remove-if-not
   (lambda (pos)
     (emacsia-block-opaque
      (emacsia-level-get-block pos)))
   coordinates)
  ;; predicate
  ;; (lambda (pos1 pos2)
  ;; 	  (let ((dx1 (- (emacsia-pos-x pos)
  ;; 			(emacsia-pos-x pos1)))
  ;; 		(dy1 (- (emacsia-pos-y pos)
  ;; 			(emacsia-pos-y pos1)))
  ;; 		(dx2 (- (emacsia-pos-x pos)
  ;; 			(emacsia-pos-x pos2)))
  ;; 		(dy2 (- (emacsia-pos-y pos)
  ;; 			(emacsia-pos-y pos2))))
  ;; 	    (< (+ (* dx1 dx1) (* dy1 dy1))
   ;; 	       (+ (* dx2 dx2) (* dy2 dy2)))))
  )

(defun emacsia-los-transform-function (origo line)
  "Return a function to transform a position to a new grid.

The new grid has LINE as it's x-axis, and ORIGO as it's origo."
  (let ((alpha (car line))
	(beta (cdr line))
	(invalpha (/ (car line) 1.0)))
    `(lambda (pos)
       ;;  new x =  distance from ORIGO intersection between line a line with a =
       ;; -1/(car line),
       ;;  new y = distance between previously mentioned intersection and
       ;;  POS.       
       (let* ((M (cons (- ,invalpha)
		       (+ (* ,invalpha (emacsia-pos-x pos))
			  (emacsia-pos-y pos))))
	      (intersection-x (/ (- (cdr M) ,beta)
				 (+ ,alpha ,invalpha)))
	      (intersection (emacsia-pos intersection-x
					 (+ (* ,alpha intersection-x)
					    ,beta)))
	      (newx (emacsia-pos-distance (quote ,origo) intersection))
	      (newy (emacsia-pos-distance pos intersection)))
	 ;; newx is now positive. that is not nececssasrily a good
	 ;; thing at _all_.
	 (emacsia-pos newx newy)))))

(defun emacsia-los-pos-between-slopes-p (playerpos blocker pos a b transform-function)
  "Return non-nil if POS is above A and below B.

A and B are cons-cells, (alpha . beta), representing line f(x) =
  alpha*x + beta."
  (let ((realpos (funcall transform-function pos)))
    (let ((yposa 0)			;(+ (* (car a) (emacsia-pos-x realpos)) (cdr a))
	  (yposb
	   (emacsia-pos-y
	    (funcall
	     transform-function
	     (emacsia-pos (emacsia-pos-x pos)
			  (+ (* (car b)
				(emacsia-pos-x pos))
			     (cdr b)))))))
      (and
       (< (emacsia-pos-y realpos) (max yposa yposb))
       (> (emacsia-pos-y realpos) (min yposa yposb))))))


;; (defun emacsia-los-pos-between-slopes-p (playerpos blocker pos a b)
;;   "Return non-nil if POS is above A and below B.

;; A and B are cons-cells, (alpha . beta), representing line f(x) =
;;   alpha*x + beta."
;;   (let ((yposa (+ (* (car a) (emacsia-pos-x pos)) (cdr a)))
;; 	(yposb (+ (* (car b) (emacsia-pos-x pos)) (cdr b)))
;; 	(realpos pos))			;(emacsia-pos-add pos (emacsia-pos 0.5 0.5))
;;     (and
;;      (< (emacsia-pos-y realpos) (max yposa yposb))
;;      (> (emacsia-pos-y realpos) (min yposa yposb)))))

(defun emacsia-los-get-ray (origin point)
  "Return a ray from ORIGIN to POINT"
  (let ((slope (/ (- (emacsia-pos-y point) (emacsia-pos-y origin))
		  (- (emacsia-pos-x point) (emacsia-pos-x origin)))))
    (cons slope (- (emacsia-pos-y origin)
		   (* slope (emacsia-pos-x origin))))))


;; Ray-corners:
;;     G
;;  A     B
;; 
;; F   #   E
;;
;;  C     D
;;     H

;; # = Opaque (and the orego)
;; A = (0.5, -0.5)  & (-0.5, 0.5)
;; B = (-0.5, -0.5) & (0.5, 0.5)
;; C = (-0.5, -0.5) & (0.5, 0.5)
;; D = (0.5, -0.5)  & (-0.5, 0.5)
;; E = (0.5, -0.5)  & (0.5, 0.5)
;; F = (-0.5, -0.5) & (-0.5, 0.5)
;; G = (-0.5, -0.5) & (0.5, -0.5)
;; H = (-0.5, 0.5) &  (0.5, 0.5)

(defun emacsia-los-lower-offset (pos1 pos2)
  "Return the offset for the lower corner of POS2 from POS1."
  (let ((offsetx 0)
	(offsety 0))
    (if (and (= (emacsia-pos-x pos1) (emacsia-pos-x pos2))
	     (< (emacsia-pos-y pos1) (emacsia-pos-y pos2)))
	(setq offsety -0.5)
      (setq offsety 0.5))
    ;; A,D & F have negative x-offsets
    (if (or
	 (and (< (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (< (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 (and (> (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (> (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 (and (= (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (< (emacsia-pos-x pos1) (emacsia-pos-x pos2))))
	(setq offsetx -0.5)
      (setq offsetx 0.5))
    (emacsia-pos offsetx offsety)))

(defun emacsia-los-upper-offset (pos1 pos2)
  "Return the offset for the uppercorner of POS2 from POS1."
  (let ((offsetx 0)
	(offsety 0))
    ;; only 'H' doesn't have -0.5 as the upper y offset
    (if (and (= (emacsia-pos-x pos1) (emacsia-pos-x pos2))
	     (> (emacsia-pos-y pos1) (emacsia-pos-y pos2)))
	(setq offsety 0.5)
      (setq offsety -0.5))
    ;; A, D, and E have positive x-offset-values
    (if (or
	 (and (< (emacsia-pos-y pos1) (emacsia-pos-y pos2))
		 (< (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 ;; D&E
	 (and (>= (emacsia-pos-y pos1) (emacsia-pos-y pos2))
		 (> (emacsia-pos-x pos1) (emacsia-pos-x pos2))))
	(setq offsetx 0.5)
      (setq offsetx -0.5))    
    (emacsia-pos offsetx offsety)))

  
(defun emacsia-los-get-upper-ray (pos1 pos2)
  "Return the ray from POS1 to upper corner of POS2."  
  ;; how to calculcate the rays, depend on which side POS is of
  ;; BLOCKER. If pos is to upperright of blocker, this works. If pos
  ;; is to the lower-right, we
  (let ((upper-corner (emacsia-pos-add pos2 (emacsia-los-upper-offset pos1 pos2)))
	(origin (emacsia-pos-add pos1 (emacsia-pos 0.0 0.0))))
    (emacsia-los-get-ray origin upper-corner)))

(defun emacsia-los-get-lower-ray (pos1 pos2)
  "Return the ray from POS1 to lower corner of POS2."
  (let ((upper-corner (emacsia-pos-add pos2 (emacsia-los-lower-offset pos1 pos2)))
	(origin (emacsia-pos-add pos1 (emacsia-pos 0.0 0.0))))
    (emacsia-los-get-ray origin upper-corner)))

(defun emacsia-los-get-shadowed-tiles (pos coordinates blocker)
  "Return tiles shadowed for POS in COORDINATES by BLOCKER."
  ;; cast two rays from POS to BLOCKER, on above, one below.
  (let* ((upper-ray (emacsia-los-get-upper-ray pos blocker))
	 (lower-ray (emacsia-los-get-lower-ray pos blocker))
	 (transform-func (emacsia-los-transform-function pos lower-ray)))
    ;; now, figure out which coordinates are being blocked.
    (remove-if-not
     (lambda (coord)
       (and (emacsia-los-pos-between-slopes-p pos blocker coord lower-ray upper-ray transform-func)
	    (> (emacsia-pos-x (funcall transform-func coord)) (emacsia-pos-x (funcall transform-func blocker)))))
     coordinates)))

;; (remove-if (lambda (pos)
;; 	     ;; remove all already visible
;; 	     ;; elements that are not
;; 	     ;; blockers.
;; 	     (and (not (equal (emacsia-level-get-visible-block pos)
;; 			      emacsia-block-unexamined))
;; 		  (not (emacsia-block-opaque (emacsia-level-get-block pos)))))
;;           (emacsia-pos-expand-radius
;; 	    pos emacsia-los-max-range))

(defun emacsia-los-highlight-from-pos-shadow (pos)
  "Makes tiles in LOS of POS visible.."
  (let ((fov-coordinates (emacsia-pos-expand-radius
			  pos emacsia-los-max-range))
	(shadowed nil))
    (let ((blockers (remove-if
		     ;; remove all blockers appearing at the edge of the map
		     (lambda (pos)
		       (or (= (emacsia-pos-y pos) (- (emacsia-level-get-height emacsia-level-current) 1))
			   (= (emacsia-pos-x pos) (- (emacsia-level-get-width emacsia-level-current) 1))
			   (= (emacsia-pos-x pos) 0)
			   (= (emacsia-pos-y pos) 0)))
		     (emacsia-los-get-blockers fov-coordinates pos))))
      (dolist (blocker blockers)
	(unless (member blocker shadowed)
	  ;; no need to check a not-visible blocker
	  (let ((newshadows (emacsia-los-get-shadowed-tiles
			     pos fov-coordinates blocker)))
	    (setq shadowed (delete-dups (append shadowed newshadows)))
	    (mapc (lambda (field)
		    (setq fov-coordinates (delq field fov-coordinates)))
		  newshadows)
	    (setq fov-coordinates (append fov-coordinates (list blocker))))))
      ;; now we have removed all shadowed fields.. return the rest
      fov-coordinates)))

(provide 'emacsia-los)
;;; emacsia-los.el ends here
