;;; emacsia-mapgen.el --- Map generating for Emacsia

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords: 

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; A map in Emacsia is a vector of strings. The vector represents
;; "lines". Each substring of length one in the strings, is actually a
;; "block", a glyph with text-properties.

;;; Code:

;; Creating maps
(defun emacsia-map-create-basic-map (width height &optional block)
  "Create a basic HEIGHT x WIDTH map.
If BLOCK is specified, use BLOCK as the basic building-block."
  (setq block (or block emacsia-block-permanent-wall))
  (let ((v (make-vector height nil)))
    (dotimes (i height v)
      (aset v i (make-vector width (emacsia-block-index block))))))

(defun emacsia-map-overwrite-region (map sx sy submap)
  "Insert SUBMAP correctly at (SX, SY) in a new copy of MAP."
  (let ((sybeg sy)
	(syend (+ sy (length submap)))
	(newmap map))
    (while (< sybeg syend)
      ;; replace this line in the new map with the corresponding in
      ;; the submap
      (let ((sxbeg sx)
	    (sxend (+ sx (length (aref submap (- sybeg sy))))))
	(while (< sxbeg sxend)
	  (setf (aref (aref newmap sybeg) sxbeg) (aref (aref submap (- sybeg sy)) (- sxbeg sx)))
	  (incf sxbeg)))
      (incf sybeg))
    ;; return the result
    newmap))

;;; mapgeneration
(defun emacsia-map-generate-room (width height &optional litp)
  "Generate a room of WIDTH x HEIGHT.

If LITP is non-nil, make the room permanently lighted."
  (let ((basemap (emacsia-map-create-basic-map width height))
	(floor (emacsia-map-create-basic-map
		(- width 2) (- height 2)
		(if litp
		    emacsia-block-floor-lit
		  emacsia-block-floor-unlit)))
	(newmap nil))
    (setq newmap (emacsia-map-overwrite-region basemap 1 1 floor))
    newmap))

(defun emacsia-map-generate-hallway (width &optional litp)
  "Generate a (left->right) hallway, of WIDTH.
if LITP is non-nil make it permanently lighted."
  (let ((basemap (emacsia-map-create-basic-map width 3 emacsia-block-permanent-wall))
	(hall (emacsia-map-create-basic-map
	       width 1
	       (if litp
		   emacsia-block-floor-lit
		 emacsia-block-floor-unlit))))
    (emacsia-map-overwrite-region basemap 0 1 hall)))

(provide 'emacsia-mapgen)
;;; emacsia-mapgen.el ends here
