;;; emacsia-pos.el --- Positioning code for Emacsia

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords:

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; 

;;; Code:

;; Consider storing positions as a single integer instead of a
;; cons-cell. This should not be too difficult, will be more memory
;; efficient, but less efficient to access individual components.

;; An idea here might be 

;; (defstruct emacsia-pos x y)

(require 'cl)

(defsubst emacsia-pos (x y)
  "Create new position with coords (X,Y)."
  (cons x y))

(defsubst emacsia-pos-x (pos)
  "Return the x-coordinate of POS."
  (car pos))

(defsubst emacsia-pos-y (pos)
  "Return the y-coordinate of POS."
  (cdr pos))

(defun emacsia-pos-p (pos)
  "Return non-nil if POS is an emacsia-pos."
  (and (integerp (car pos)) (integerp (cdr pos))))

(defun emacsia-pos-dot-product (pos1 pos2)
  "Return the dot-product of POS1 and POS2, as vectors."
  (+ (* (emacsia-pos-x pos1) (emacsia-pos-x pos2))
     (* (emacsia-pos-y pos1) (emacsia-pos-y pos2))))

(defun emacsia-pos-distance-predicate-p (pos1 pos2 distance pred)
  "Call PRED on distance(POS1,POS2) and DISTANCE."
  (let ((dx (- (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	(dy (- (emacsia-pos-y pos1) (emacsia-pos-y pos2))))
    (funcall pred (+ (* dx dx) (* dy dy)) (* distance distance))))

(defsubst emacsia-pos-distance (pos1 pos2)
  "Distance between pos1 and pos2"
  (emacsia-pos-length (emacsia-pos-subtract pos2 pos1)))

(defsubst emacsia-pos-length (pos)
  "Return the length of POS as a vector."
  (sqrt (+ (* (emacsia-pos-x pos) (emacsia-pos-x pos))
	   (* (emacsia-pos-y pos) (emacsia-pos-y pos)))))

(define-hash-table-test 'emacsia-arccos-test 'emacsia-arccos-equals-p 'sxhash)

(defvar emacsia-arccos-lookup-table
  (make-hash-table :test 'emacsia-arccos-test)
  "A hash-table of arc-cos values used for fast lookup.")

(defvar emacsia-arccos-epsilon 0.001
  "The amount of error allowed in a number.")

(defun emacsia-arccos-equals-p (val1 val2)
  "Return non-nil if VAL1 and VAL2 represent the same number."
  (> emacsia-arccos-epsilon (abs (- val2 val1))))

(defun emacsia-arccos (val)
  "Return the acos of val."
  (or (gethash val emacsia-arccos-lookup-table)
      (setf (gethash val emacsia-arccos-lookup-table)
	    (acos val))))

(defun emacsia-arccos-cutoff (x)
  (if (> x 1.0)
      1.0
    (if (< x -1.0)
	1.0
      x)))

(defun emacsia-pos-angle-between (vec1 vec2)
  "Return the angle between VEC1 and VEC2 in degrees."
  (acos
   (emacsia-arccos-cutoff (/ (emacsia-pos-dot-product vec1 vec2)
			     (emacsia-pos-length vec1)
			     (emacsia-pos-length vec2)))))

(defvar emacsia-pos-radius-expansion nil
  "A vector of relative radius expansion coordinate lists.")

(defun emacsia-pos-calculate-relative-expansion (radius)
  "Calculate a relative expansion list."
  (let ((result nil)
	(x (- radius))
	(y (- radius)))
    (while (<= x radius)
      (setq y (- radius))
      (while (<= y radius)
	;; max(x, y) + (min(x, y) / 2) <= radius
	(when (<= (+ (* x x)
		     (* y y))
		  (* radius radius))
;; 	    (<= (+ (max x y) (/ (min x y) 2)) radius)
	  (add-to-list 'result (emacsia-pos x y)))
	(setq y (+ y 1)))
      (setq x (+ x 1)))
    result))

(defun emacsia-pos-get-relative-expansion (radius)
  "Get a relative expansion of RADIUS."
  (unless emacsia-pos-radius-expansion
    ;; twelve entries should be enough for now.
    (setq emacsia-pos-radius-expansion (make-vector 12 nil)))
  (if (aref emacsia-pos-radius-expansion radius)
      ;; return it
      (aref emacsia-pos-radius-expansion radius)
    ;; else build, save and return it.
    (let ((coords (emacsia-pos-calculate-relative-expansion
		   radius)))
      (aset emacsia-pos-radius-expansion radius coords)
      coords)))

;; Simple arithmetics
(defun emacsia-pos-add (pos1 pos2)
  "Add POS1 and POS2 and return the result."
  (emacsia-pos (+ (emacsia-pos-x pos1)
		       (emacsia-pos-x pos2))
		    (+ (emacsia-pos-y pos1)
		       (emacsia-pos-y pos2))))

(defun emacsia-pos-subtract (pos1 pos2)
  "Substract POS2 from POS1 and return the result."
  (emacsia-pos (- (emacsia-pos-x pos1)
		  (emacsia-pos-x pos2))
	       (- (emacsia-pos-y pos1)
		  (emacsia-pos-y pos2))))

(defun emacsia-pos-multiply (pos1 factor)
  "Multiply POS1 with FACTOR and return the result."
  (emacsia-pos (* (emacsia-pos-x pos1) factor)
	       (* (emacsia-pos-y pos1) factor)))

(defun emacsia-pos-normalize (vector)
  "Normalize VECTOR to have length 1."
  (emacsia-pos-multiply vector (/ 1.0 (emacsia-pos-length vector))))

(defun emacsia-pos-expand-radius (pos radius)
  "Expand POS in RADIUS, and return all coordiantes in a list."
  ;; as a first beta, let's just expand in a cross.
  (let ((coords `(,pos)))
    (dolist (relpos (emacsia-pos-get-relative-expansion radius))
      (setq coords (append coords
			   (list (emacsia-pos-add pos relpos)))))
    (let ((h (emacsia-level-get-height emacsia-level-current))
	  (w (emacsia-level-get-width  emacsia-level-current)))
      (remove-if (lambda (coord)
		   (or (< (emacsia-pos-x coord) 0)
		       (>= (emacsia-pos-x coord) w)
		       (< (emacsia-pos-y coord) 0)
		       (>= (emacsia-pos-y coord) h)))
		 coords))))

(defun emacsia-pos-get-touching (pos)
  "Return all positions touching POS."
  (mapcar (lambda (dir) (emacsia-pos-add pos dir))
	  emacsia-directions))

(defconst emacsia-direction-north (emacsia-pos 0 -1)
  "Direction vector.")

(defconst emacsia-direction-south (emacsia-pos 0 1)
  "Direction vector.")

(defconst emacsia-direction-east (emacsia-pos 1 0)
  "Direction vector.")

(defconst emacsia-direction-west (emacsia-pos -1 0)
  "Direction vector.")

(defconst emacsia-direction-northeast (emacsia-pos 1 -1)
  "Direction vector.")

(defconst emacsia-direction-northwest (emacsia-pos -1 -1)
  "Direction vector.")

(defconst emacsia-direction-southeast (emacsia-pos 1 1)
  "Direction vector.")

(defconst emacsia-direction-southwest (emacsia-pos -1 1)
  "Direction vector.")

(defconst emacsia-directions (list emacsia-direction-west
				   emacsia-direction-south
				   emacsia-direction-southwest
				   emacsia-direction-southeast
				   emacsia-direction-north
				   emacsia-direction-northwest
				   emacsia-direction-northeast
				   emacsia-direction-east)
  "A list of all the different directions.")

(provide 'emacsia-pos)
;;; emacsia-pos.el ends here
