;;; emacsia-shadows.el --- Shadowing-LOS for Emacsia

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords: 

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; Provide

;;; Code:
(require 'cl)

(defvar emacsia-shadow-max-range 11
  "The max LOS range.")

(defsubst emacsia-shadow-same-side-cp (vec1 vec2)
  "Return the Z-component of VEC1 cross VEC2.
VEC1 and VEC2 should both be vectors in the xy-plane."
  (- (* (emacsia-pos-x vec1) (emacsia-pos-y vec2))
     (* (emacsia-pos-y vec1) (emacsia-pos-x vec2))))

(defsubst emacsia-shadow-same-side-dp (vec1 vec2)
  "Return the dot-product of two vectors with Z-coordinates VEC1 and VEC2.
This is a special dot-product for floats representing the
z-coordinate of a crossproduct of two vectors in the xy-plane"  
  (+ (* vec1 vec2)))

(defsubst emacsia-shadow-same-side (point refpoint linefrom lineto)
  "Deternmine if POINT is on the same side of LINE as REFPOINT."  
  (>= (emacsia-shadow-same-side-dp
       (emacsia-shadow-same-side-cp
	(emacsia-pos-subtract lineto linefrom)
	(emacsia-pos-subtract point linefrom))
       (emacsia-shadow-same-side-cp
	(emacsia-pos-subtract lineto linefrom)
	(emacsia-pos-subtract refpoint linefrom)))
      -0.001))

(defun emacsia-shadow-point-in-polygon (point polygon)
  "Return true if POINT is inside the quadrangle POLYGON."
  (and (emacsia-shadow-same-side point (aref polygon 0)
				 (aref polygon 1) (aref polygon 2))
       (emacsia-shadow-same-side point (aref polygon 1)
				 (aref polygon 2) (aref polygon 3))
       (emacsia-shadow-same-side point (aref polygon 2)
				 (aref polygon 3) (aref polygon 0))
       (emacsia-shadow-same-side point (aref polygon 3)
				 (aref polygon 0) (aref polygon 1))))

;; Old bad version. Angle summation is *the* slowest point-in-poly
;; algo.

;; (defun emacsia-shadow-point-in-poly (point poly)
;;   "Return non-nil if POINT is inside POLY."
;;   (let ((sum 0.0)
;; 	(firstvertex (car poly)))
;;     (while poly
;;       (if (cdr poly)
;; 	  (setq sum (+ sum (emacsia-pos-angle-between
;; 			    (emacsia-pos-subtract (car poly) point)
;; 			    (emacsia-pos-subtract (cadr poly) point))))
;; 	(setq sum (+ sum (emacsia-pos-angle-between
;; 			  (emacsia-pos-subtract (car poly) point)
;; 			  (emacsia-pos-subtract firstvertex point)))))
;;       (setq poly (cdr poly)))
;;     (emacsia-arccos-equals-p sum 6.28318)))

(defun emacsia-shadow-get-poly (pos blocker)
  "Return the blocked polygon of POS by BLOCKER."
  (let ((upper-corner (emacsia-pos-add (emacsia-shadow-upper-offset pos blocker) blocker))
	(lower-corner (emacsia-pos-add (emacsia-shadow-lower-offset pos blocker) blocker))
	(blockx (emacsia-pos-x blocker)))
    (vector upper-corner
	    lower-corner
	    (emacsia-pos-add (emacsia-pos-multiply
			      (emacsia-pos-normalize
			       (emacsia-pos-subtract lower-corner pos))
			      (* emacsia-shadow-max-range 2))
			     lower-corner)	    
	    (emacsia-pos-add (emacsia-pos-multiply
			      (emacsia-pos-normalize
			       (emacsia-pos-subtract
				upper-corner pos))
			      (* emacsia-shadow-max-range 2))
			     upper-corner))))

(defun emacsia-shadow-get-shadowed (pos blocker coordinates)
  "Return the COORDINATES blocked for POS by BLOCKER."
  (let ((blockpoly (emacsia-shadow-get-poly pos blocker)))
    (delete blocker (remove-if-not (lambda (pos)
				   (emacsia-shadow-point-in-polygon pos blockpoly))
				 coordinates))))

(defun emacsia-shadow-get-all-shadowed (pos)
  "Return all visible tiles from POS."
  (let ((fov-coordinates (emacsia-pos-expand-radius
			  pos emacsia-shadow-max-range))
	(shadowed nil))
    (let ((blockers (remove-if
		     ;; remove all blockers appearing at the edge of the map
		     (lambda (pos)
		       (or (= (emacsia-pos-y pos) (- (emacsia-level-get-height emacsia-level-current) 1))
			   (= (emacsia-pos-x pos) (- (emacsia-level-get-width emacsia-level-current) 1))
			   (= (emacsia-pos-x pos) 0)
			   (= (emacsia-pos-y pos) 0)))
		     (emacsia-los-get-blockers fov-coordinates pos))))
      (dolist (blocker blockers)
	(unless (member blocker shadowed)
	  ;; no need to check a not-visible blocker
	  (let ((newshadows (emacsia-shadow-get-shadowed 
			     pos blocker fov-coordinates)))
	    (setq shadowed (delete-dups (append shadowed newshadows)))
	    (mapc (lambda (field)
		    (setq fov-coordinates (delete field fov-coordinates)))
		  newshadows))))  ;; (setq fov-coordinates (append fov-coordinates (list blocker)))      
      ;; now we have removed all shadowed fields.. return the rest
      fov-coordinates)))

(defun emacsia-shadow-lower-offset (pos1 pos2)
  "Return the offset for the lower corner of POS2 from POS1."
  (let ((offsetx 0)
	(offsety 0))
    (if (and (= (emacsia-pos-x pos1) (emacsia-pos-x pos2))
	     (< (emacsia-pos-y pos1) (emacsia-pos-y pos2)))
	(setq offsety -0.5)
      (setq offsety 0.5))
    ;; A,D & F have negative x-offsets
    (if (or
	 (and (< (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (< (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 (and (> (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (> (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 (and (= (emacsia-pos-y pos1) (emacsia-pos-y pos2))
	      (< (emacsia-pos-x pos1) (emacsia-pos-x pos2))))
	(setq offsetx -0.5)
      (setq offsetx 0.5))
    (emacsia-pos offsetx offsety)))

(defun emacsia-shadow-upper-offset (pos1 pos2)
  "Return the offset for the uppercorner of POS2 from POS1."
  (let ((offsetx 0)
	(offsety 0))
    ;; only 'H' doesn't have -0.5 as the upper y offset
    (if (and (= (emacsia-pos-x pos1) (emacsia-pos-x pos2))
	     (> (emacsia-pos-y pos1) (emacsia-pos-y pos2)))
	(setq offsety 0.5)
      (setq offsety -0.5))
    ;; A, D, and E have positive x-offset-values
    (if (or
	 (and (< (emacsia-pos-y pos1) (emacsia-pos-y pos2))
		 (< (emacsia-pos-x pos1) (emacsia-pos-x pos2)))
	 ;; D&E
	 (and (>= (emacsia-pos-y pos1) (emacsia-pos-y pos2))
		 (> (emacsia-pos-x pos1) (emacsia-pos-x pos2))))
	(setq offsetx 0.5)
      (setq offsetx -0.5))    
    (emacsia-pos offsetx offsety)))
  
(provide 'emacsia-shadows)
;;; emacsia-shadows.el ends here
