;;; emacsia.el --- Emacsia

;; Copyright (C) 2004  Free Software Foundation, Inc.

;; Author: Ulrik Jensen <terryp@daimi.au.dk>
;; Keywords: 

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; 

;;; Code:
(require 'emacsia-pos)
(require 'emacsia-level)
(require 'emacsia-block)
(require 'emacsia-mapgen)
(require 'emacsia-gg)
(require 'emacsia-los)
(require 'emacsia-shadows)

(require 'cl)

(defvar emacsia-touched nil
  "A list of fields that have been touched by Emacsia in the last.")

(defvar emacsia-playing-p nil
  "Non-nil if a game is currently being played.")

(defvar emacsia-current-position nil
  "The players current position.")

(defvar emacsia-step-energy nil
  "The amount of energy required to take a step.")

(defun emacsia-touch (pos)
  "Mark POS as updated.

POS should either be an `emacsia-pos' or a list of them."
  (if (emacsia-pos-p pos)
      (add-to-list 'emacsia-touched pos)
    (mapc 'emacsia-touch pos)))

(defvar emacsia-player-light-radius 3
  "A temporary variable for the players light-radius.")

(defun emacsia-gamecycle ()
  "Perform a single gamecycle."
  ;; if the player has enough energy to move, allow him to do so.
  ;; move all monsters with enough energy.
  ;; monsters could have moved, apply LOS
  
  ;; update display
  (emacsia-gg-update))

(defun emacsia-move-possible-p (dest)
  "Return non-nil if dest can be moved to by the player."
  ;; currently just handle obstacles
  (not (emacsia-block-obstaclep
	(emacsia-level-get-block dest))))

(defun emacsia-use-energy (amount)
  "Use an AMOUNT of the players energy"
  ;; Use an amount of energy. If not enough is left to move, call
  ;; cycle. Otherwise just contineue.
  (emacsia-gamecycle))

(defvar emacsia-current-los-fields nil
  "A list of the fields currently in LOS.")

(defun emacsia-move (direction)
  "Move the player."
  ;; handle speed/energy
  (let ((dest (emacsia-pos-add emacsia-current-position
			       direction)))
    (if (not (emacsia-move-possible-p dest))
	(error "Not a possible move!")
      ;; do the move
      (emacsia-touch emacsia-current-position)
      ;; touch the old lighting radius, in case the underlying tiles
      ;; were unlighted.
      (when emacsia-current-los-fields
	(mapc (lambda (elm)
		(emacsia-level-unlight elm))
	      emacsia-current-los-fields))
      ;; remove all fields that are not either permalit or within
      ;; light-radius!
      (let ((los-fields ;; (emacsia-los-highlight-from-pos-shadow dest)
	     (emacsia-shadow-get-all-shadowed dest)))
	(setq emacsia-current-los-fields
	      ;; highlight fields that are ((in lightradius && in los) || (in los && permalit))
	      los-fields))
      (setq emacsia-current-position dest)
      (when emacsia-current-los-fields
	(mapc (lambda (pos)
		(when (or (emacsia-pos-distance-predicate-p dest pos emacsia-player-light-radius '<=)
			  (emacsia-block-permalit (emacsia-level-get-block pos)))
		  (emacsia-level-light pos)))
	      emacsia-current-los-fields))
      ;; add lighting effects -- fixed lighting radius currently.
      (emacsia-use-energy emacsia-step-energy))))

;; (union (intersection los-fields light-fields :test 'equal)
;; 		     (remove-if-not
;; 		      (lambda (pos)
;; 			(emacsia-block-permalit (emacsia-level-get-block pos)))
;; 		      los-fields :test 'equal))

(defun emacsia-turn ()
  "Perform one game turn."
  ;; assign new energy to everything
  ;; while something has enough energy to do something, let it move.
  )

(defun emacsia-start ()
  "Start a game of Emacsia."
  (interactive)
  ;; Create or load a character
  (setq emacsia-current-position (emacsia-pos 1 1))
  ;; Load blocks
  (emacsia-block-load-blocks)
  ;; Set up level
  (let ((map (emacsia-map-create-basic-map 40 20 emacsia-block-unexamined)))
    (setq map (emacsia-map-overwrite-region map 0 0 (emacsia-map-generate-room 20 20 nil)))
    (setq map (emacsia-map-overwrite-region map 13 13 (emacsia-map-generate-room 7 7 t)))
    (setq map (emacsia-map-overwrite-region map 19 16 (emacsia-map-generate-hallway 7 nil)))
    (setq map (emacsia-map-overwrite-region map 26 9 (emacsia-map-generate-room 10 10 t)))
    (emacsia-set-level
     (make-emacsia-level :name "Debugging"
			 :depth 50
			 :map map
			 :visible
			 (emacsia-map-create-basic-map
			  40 20
			  emacsia-block-unexamined)))
    (emacsia-level-set-block (emacsia-pos 6 6) emacsia-block-permanent-wall)
    (emacsia-level-set-block (emacsia-pos 6 7) emacsia-block-permanent-wall)
    (emacsia-level-set-block (emacsia-pos 6 8) emacsia-block-permanent-wall)
    ;; generate a bigger wall with a hole in it.
    (emacsia-level-set-block (emacsia-pos 6 12) emacsia-block-permanent-wall)
    (emacsia-level-set-block (emacsia-pos 6 13) emacsia-block-permanent-wall)
    (emacsia-level-set-block (emacsia-pos 6 15) emacsia-block-permanent-wall)
    (emacsia-level-set-block (emacsia-pos 6 16) emacsia-block-permanent-wall)
    ;; generate a door for the smaller room
    (emacsia-level-set-block (emacsia-pos 16 13) emacsia-block-door-opened)
    ;; and the door into the last room
    (emacsia-level-set-block (emacsia-pos 26 17) emacsia-block-door-opened))
  ;; Initialise UI
  (emacsia-gg-init-buffer)  
  (setq emacsia-playing-p t)
  (emacsia-move emacsia-direction-southeast))

(provide 'emacsia)
;;; emacsia.el ends here
